// Andreas Kirsch 2007

#include "stdlib.h"
#include "memory.h"
#include <iostream>

// More Advanced Coroutine Implementation
#define COROUTINE( returnType, name, paramList ) class name { \
   size_t yieldOffset; \
   void *stackCopy; \
   size_t stackSize; \
   \
public: \
   name() : stackCopy( 0 ), stackSize( 0 ), yieldOffset( 0 ) { \
   } \
   \
   ~name() { \
   /* this is something bad actually - it should always finish but oh well */ \
      if( stackCopy ) { \
         delete stackCopy; \
      } \
   } \
   \
   returnType operator() paramList { \
      size_t __yieldExitOffset; \
      __asm { mov __yieldExitOffset, offset yieldExit_##name } \
      returnType __yieldReturnValue; \
      char __frameStart; \
      \
      if( stackCopy ) { \
         /* restore the stack */ \
         memcpy( &__frameStart - stackSize, stackCopy, stackSize ); \
         size_t localYieldOffset = yieldOffset; \
         __asm { jmp localYieldOffset } \
      }

#define CONCAT( a, b ) a ## b
#define LABEL( line ) CONCAT( __label__, line )

#define YIELD( value ) do { \
      char __frameEnd; \
      /* copy the stack frame */ \
      if( stackCopy ) { \
         delete stackCopy; \
      } \
      stackSize = &__frameStart - &__frameEnd; \
      stackCopy = new char[ stackSize ]; \
      memcpy( stackCopy, &__frameEnd, stackSize ); \
      \
      __yieldReturnValue = (value); \
      size_t localYieldOffset; \
      __asm { mov localYieldOffset, offset LABEL( __LINE__ ) } \
      yieldOffset = localYieldOffset; \
      __asm { jmp __yieldExitOffset } \
LABEL( __LINE__ ): \
      ; \
   } while( 0 )

#define RETURN( value ) \
   do {\
      /* remove the stack copy and clean up */ \
      if( stackCopy ) { \
         delete stackCopy; \
         stackCopy = 0; \
         stackSize = 0; \
      } \
      \
      yieldOffset = 0; \
      return (value);\
   } while( 0 )

#define ENDCOROUTINE( name ) \
         yieldExit_##name: \
         return __yieldReturnValue; \
      } /* function done */ \
   }; /* class done */

// Example Code
COROUTINE( int, testCoroutine, () ) {
   std::cout << "Hello World" << std::endl;
   YIELD( 1 );
   std::cout << "Again: I'm back!" << std::endl;
   YIELD( 2 );
   std::cout << "And one last time!" << std::endl;
   RETURN( 3 );
} ENDCOROUTINE( testCoroutine )

COROUTINE( int, fibonnacci, () ) {
   int a = 1, b = 1;

   while( 1 ) {
      YIELD( b );
      int c = a + b;
      a = b;
      b = c;
   }
} ENDCOROUTINE( fibonnacci )

void testAdvancedImplementation() {
   testCoroutine testCoroutine;
   testCoroutine();
   testCoroutine();
   testCoroutine();
   // now it starts from the beginning again
   testCoroutine();

   fibonnacci fibonnacciTest;
   fibonnacci fibonnacciTest2;
   for( int i = 0 ; i < 10 ; i++ ) {
      std::cout << "First: " << fibonnacciTest() << std::endl;
      if( i % 2 == 0 ) {
         std::cout << "Second: " << fibonnacciTest2() << std::endl;
      }
   }
}