// Andreas Kirsch 2007

#include <iostream>

// Simple Coroutine Implementation
#define YIELD( value ) \
   do { \
      yieldIndex = __COUNTER__; \
      return (value); \
      case __COUNTER__ - 1: \
         ; \
   } while( 0 )

#define RETURN( value ) \
   do { \
      yieldIndex = -1; \
      return (value); \
   } while( 0 )

#define COROUTINE( returnType, name ) \
   returnType name() { \
      static int yieldIndex = -1; \
      switch( yieldIndex ) { \
            case -1:

#define ENDCOROUTINE \
         yieldIndex = -1; \
      } \
   }

// Examples
COROUTINE( int, simpleCoroutine ) {
   static int a;
   a = 0;
   while( 1 ) {
      a++;
      YIELD( a );
   }
} ENDCOROUTINE

COROUTINE( int, simpleForLoop ) {
   static int a;
   for( a = 0 ; a < 5 ; a++ ) {
      static int b;
      for( b = 0 ; b < 5 ; b++ ) {
         YIELD( a * 10 + b );
         if( a == b ) {
            YIELD( -2 );
         }
      }
   }
   RETURN( -1 );
} ENDCOROUTINE

COROUTINE( unsigned, simpleFibbonacci ) {
   static unsigned a, b;
   a = 0;
   b = 1;
   while( 1 ) {
      static unsigned c;

      YIELD( b );
      c = a;
      a = b;
      b += c;
   }
} ENDCOROUTINE

COROUTINE( const char *, simpleMonolog ) {
   YIELD( "Hello.\n" );
   YIELD( "I'm a coroutine\n" );
   YIELD( "I can yield a return value and next time I'm called, I'll continue where I have left off.\n" );
   YIELD( "This is great and is really easy!\n" );
   RETURN( "And when I'm done I'll start from the beginning next time.\nBye!\n" );
} ENDCOROUTINE

void testSimpleImplementation() {
   int i;
   for( i = 0 ; i < 5 ; i++ ) {
      std::cout << simpleMonolog();
   }
   
   int result;   
   while( (result = simpleForLoop()) != -1 ) {
      std::cout << result << std::endl;
   }

   for( i = 0 ; i < 10 ; i++ ) {
      std::cout << simpleFibbonacci() << std::endl;
   }
}
